#############################################################################
##
#W  test.gd                 GAP library                          Frank Lübeck
##
#Y  Copyright (C) 2011 The GAP Group
##
##  This file declares functions for test files.
##

DeclareGlobalFunction("ParseTestInput");
DeclareGlobalFunction("ParseTestFile");
DeclareGlobalFunction("RunTests");
DeclareGlobalFunction("Test");
DeclareGlobalFunction("TestDirectory");
DeclareGlobalFunction("TestPackage");

