/****************************************************************************
**
*W  compstat.h                  GAP source                       Frank Celler
**
**
*Y  Copyright (C)  1996,  Lehrstuhl D für Mathematik,  RWTH Aachen,  Germany
*Y  (C) 1998 School Math and Comp. Sci., University of St Andrews, Scotland
*Y  Copyright (C) 2002 The GAP Group
*/

#ifndef GAP_COMPSTAT_H
#define GAP_COMPSTAT_H

/****************************************************************************
**
*V  CompInitFuncs . . . . . . . . . .  list of compiled module init functions
**
**  This a dummy list in case no module is statically linked.
*/
extern InitInfoFunc CompInitFuncs [];


#endif // GAP_COMPSTAT_H
