/****************************************************************************
**
*W  sctable.h                   GAP source                     Marcel Roelofs
**
**
*Y  Copyright (C)  1996,        CWI,        Amsterdam,        The Netherlands
*Y  (C) 1998 School Math and Comp. Sci., University of St Andrews, Scotland
*Y  Copyright (C) 2002 The GAP Group
**
**  This file declares a fast access  function for structure constants tables
**  and the multiplication of two elements using a structure constants table.
*/

#ifndef GAP_SCTABLE_H
#define GAP_SCTABLE_H

#include "system.h"

/****************************************************************************
**
*F * * * * * * * * * * * * * initialize module * * * * * * * * * * * * * * *
*/

/****************************************************************************
**
*F  InitInfoSCTable() . . . . . . . . . . . . . . . . table of init functions
*/
StructInitInfo * InitInfoSCTable ( void );


#endif // GAP_SCTABLE_H
