/****************************************************************************
**
*W  costab.h                    GAP source                       Frank Celler
*W                                                           & Volkmar Felsch
*W                                                         & Martin Schönert
**
**
*Y  Copyright (C)  1996,  Lehrstuhl D für Mathematik,  RWTH Aachen,  Germany
*Y  (C) 1998 School Math and Comp. Sci., University of St Andrews, Scotland
*Y  Copyright (C) 2002 The GAP Group
**
**  This file contains the functions for coset tables.
*/

#ifndef GAP_COSTAB_H
#define GAP_COSTAB_H

#include "system.h"

/****************************************************************************
**
*F * * * * * * * * * * * * * initialize module * * * * * * * * * * * * * * *
*/

/****************************************************************************
**
*F  InitInfoCosetTable() . . . . . . . . . . . . . .  table of init functions
*/
StructInitInfo * InitInfoCosetTable ( void );



#endif // GAP_COSTAB_H
