/****************************************************************************
**
*W  tietze.h                    GAP source                     Volkmar Felsch
**
**
*Y  Copyright 1990-1992,  Lehrstuhl D für Mathematik,  RWTH Aachen,  Germany
*Y  (C) 1998 School Math and Comp. Sci., University of St Andrews, Scotland
*Y  Copyright (C) 2002 The GAP Group
**
**  This file defines the functions for computing with finite presentations.
*/

#ifndef GAP_TIETZE_H
#define GAP_TIETZE_H

#include "system.h"

/****************************************************************************
**
*F * * * * * * * * * * * * * initialize module * * * * * * * * * * * * * * *
*/

/****************************************************************************
**
*F  InitInfoTietze()  . . . . . . . . . . . . . . . . table of init functions
*/
StructInitInfo * InitInfoTietze ( void );


#endif // GAP_TIETZE_H
